#!/usr/bin/python3
import apt
import apt_pkg
import logging
import unittest

import sys
sys.path.insert(0, "../")

from LanguageSelector.LocaleInfo import LocaleInfo
from LanguageSelector.LangCache import LanguageSelectorPkgCache, LanguageInformation

class TestLangCache(unittest.TestCase):

    def setUp(self):
        apt_pkg.config.set("Dir::State::lists","./test-data/var/lib/apt/lists.cl")
        apt_pkg.config.set("Dir::State::status","./test-data/var/lib/dpkg/status")
        apt_pkg.config.set("Dir::Etc::SourceList","./test-data/etc/apt/sources.list.cl")
        apt_pkg.config.set("Dir::Etc::SourceParts","x")
        logging.info("updating the cache")
        localeinfo = LocaleInfo("languagelist", "..")
        self.lang_cache = LanguageSelectorPkgCache(
            localeinfo, apt.progress.base.OpProgress())
        self.lang_cache.update()
        self.lang_cache.open()

    def test_get_language_information(self):
        """ test if getLanguagenformation returns values """
        language_info = self.lang_cache.getLanguageInformation()
        self.assertTrue(len(language_info) > 1)
        self.assertEqual(len([x for x in language_info if x.languageCode == "de"]), 1)

    def test_lang_info(self):
        """ test if tryChangeDetails works """
        self.assertEqual(len(self.lang_cache.get_changes()), 0)
        # create LanguageInformation object and test basic properties
        li = LanguageInformation(self.lang_cache, "de", "german")
        self.assertFalse(li.languagePkgList["languagePack"].installed)
        self.assertTrue(li.languagePkgList["languagePack"].available)

    def test_try_change_details(self):
        li = LanguageInformation(self.lang_cache, "de", "german")
        # test if writing aids get installed 
        li.languagePkgList["languagePack"].doChange = True
        self.lang_cache.tryChangeDetails(li)
        self.assertTrue(self.lang_cache["language-pack-de"].marked_install)
        self.assertTrue(self.lang_cache["hyphen-de"].marked_install)

if __name__ == "__main__":
    apt_pkg.config.set("Apt::Architecture","i386")
    unittest.main()



#!/usr/bin/python3
# coding: UTF-8

import unittest
import os.path
import os
import sys
import apt
import tempfile
import shutil
import resource
import copy

srcdir = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))
sys.path.insert(0, srcdir)
from language_support_pkgs import LanguageSupport, DEFAULT_DEPENDS_FILE, apt_cache_add_language_packs

class T(unittest.TestCase):
    @classmethod
    def setUpClass(klass):
        # initialize this once for better performance
        klass.apt_cache = apt.Cache()
        klass.pkg_depends = os.path.join(srcdir, 'data', 'pkg_depends')

    def setUp(self):
        # in case tests modify $PATH, save it here and restore it in tearDown
        self.orig_path = os.getenv('PATH')
        self.workdir = tempfile.mkdtemp()

    def tearDown(self):
        if self.orig_path:
            os.environ['PATH'] = self.orig_path
        shutil.rmtree(self.workdir)

    @unittest.skipUnless(os.path.exists(DEFAULT_DEPENDS_FILE), 
            'no system installed pkg_depends file')
    def test_parse_pkg_depends_system(self):
        '''Parse system-installed pkg_depends file'''

        ls = LanguageSupport(self.apt_cache)
        self.assertGreater(len(ls.pkg_depends), 5)

    def test_parse_pkg_depends_local(self):
        '''Parse pkg_depends file in source tree'''

        ls = LanguageSupport(self.apt_cache, self.pkg_depends)
        self.assertGreater(len(ls.pkg_depends), 5)
        self.assertTrue('de' in ls.pkg_depends[''])
        self.assertTrue('' in ls.pkg_depends['firefox'])
        self.assertTrue('tr' in ls.pkg_depends['firefox'][''])
        self.assertTrue('wa' in ls.pkg_depends['']['de'])

    def test_langcode_from_locale(self):
        '''_langcode_from_locale()'''

        self.assertEqual(LanguageSupport._langcode_from_locale('de'), 'de')
        self.assertEqual(LanguageSupport._langcode_from_locale('de_DE.UTF-8'), 'de')
        self.assertEqual(LanguageSupport._langcode_from_locale('en_GB'), 'en')
        self.assertEqual(LanguageSupport._langcode_from_locale('be_BY@latin'), 'be')
        self.assertEqual(LanguageSupport._langcode_from_locale('zh_CN.UTF-8'), 'zh-hans')
        self.assertEqual(LanguageSupport._langcode_from_locale('zh_TW'), 'zh-hant')

    def test_by_package_and_locale_trigger(self):
        '''by_package_and_locale() for a trigger package'''

        ls = LanguageSupport(self.apt_cache, self.pkg_depends)

        result = ls.by_package_and_locale('libreoffice-common', 'es_ES.UTF-8', True)
        self._check_valid_pkgs(result)
        # implicit locale suffix
        self.assertTrue('libreoffice-l10n-es' in result)
        self.assertTrue('libreoffice-help-es' in result)
        # explicit entry for that language
        self.assertTrue('myspell-es' in result)

        # language only
        result = ls.by_package_and_locale('libreoffice-common', 'de', True)
        self._check_valid_pkgs(result)
        self.assertTrue('libreoffice-l10n-de' in result)
        self.assertTrue('libreoffice-help-de' in result)

        # Chinese special case
        result = ls.by_package_and_locale('firefox', 'zh_CN.UTF-8', True)
        self._check_valid_pkgs(result)
        self.assertTrue('firefox-locale-zh-hans' in result)
        # no generic packages
        self.assertFalse('language-pack-zh-hans' in result)

        result = ls.by_package_and_locale('libreoffice-common', 'zh_CN.UTF-8', True)
        self._check_valid_pkgs(result)
        self.assertTrue('libreoffice-l10n-zh-cn' in result)
        self.assertTrue('libreoffice-help-zh-cn' in result)

        # without locale suffix
        result = ls.by_package_and_locale('chromium-browser', 'dv_MV', True)
        self._check_valid_pkgs(result)
        self.assertTrue('chromium-browser-l10n' in result)

    def test_by_package_and_locale_generic(self):
        '''by_package_and_locale() for generic support'''

        ls = LanguageSupport(self.apt_cache, self.pkg_depends)

        result = ls.by_package_and_locale('', 'en_GB.UTF-8', True)
        self._check_valid_pkgs(result)
        self.assertTrue('language-pack-en' in result)
        self.assertTrue('wbritish' in result)

        # language code only
        result = ls.by_package_and_locale('', 'de', True)
        self._check_valid_pkgs(result)
        self.assertTrue('language-pack-de' in result)
        self.assertTrue('wngerman' in result)

    def test_by_package_and_locale_noinstalled(self):
        '''by_package_and_locale() without installed packages'''

        ls = self._fake_apt_language_support(['language-pack-en'],
                ['wbritish'])

        result = ls.by_package_and_locale('', 'en_GB.UTF-8', False)
        self.assertEqual(result, ['wbritish'])

        ls = self._fake_apt_language_support(['language-pack-en', 'firefox',
            'firefox-locale-en', 'libreoffice-common', 'gedit'], 
            ['wbritish', 'firefox-locale-uk', 'hunspell-en-ca',
             'hunspell-en-us', 'aspell-en'])
        result = ls.by_package_and_locale('', 'en_GB.UTF-8', False)
        self.assertEqual(result, ['wbritish'])
        result = ls.by_package_and_locale('firefox', 'en_GB.UTF-8', False)
        self.assertEqual(result, [])
        result = ls.by_package_and_locale('abiword', 'en_GB.UTF-8', False)
        self.assertEqual(result, ['aspell-en'])

        # if we specify a country, do not give us stuff for other countries
        result = ls.by_package_and_locale('libreoffice-common', 'en_GB.UTF-8', False)
        self.assertEqual(result, [])

        # if we specify just the language, give us stuff for all countries
        result = ls.by_package_and_locale('libreoffice-common', 'en', False)
        self.assertEqual(set(result), set(['hunspell-en-ca', 'hunspell-en-us']))

    def test_by_package_and_locale_unknown(self):
        '''by_package_and_locale() for unknown locales/triggers'''

        ls = LanguageSupport(self.apt_cache, self.pkg_depends)

        result = ls.by_package_and_locale('unknown_pkg', 'de_DE.UTF-8', True)
        self.assertEqual(result, [])
        result = ls.by_package_and_locale('firefox', 'bo_GUS', True)
        self.assertEqual(result, [])

    def test_by_locale(self):
        '''by_locale()'''

        ls = self._fake_apt_language_support(['firefox-locale-de',
            'language-pack-de', 'gvfs', 'libreoffice-common', 'firefox'],
            ['language-pack-gnome-de', 'language-pack-kde-de', 'wngerman',
            'wbritish', 'libreoffice-help-de'])
        result = ls.by_locale('de_DE.UTF-8', True)
        self.assertEqual(set(result), set(['firefox-locale-de', 'language-pack-de',
            'language-pack-gnome-de', 'wngerman', 'libreoffice-help-de']))

        # no duplicated items in result list
        self.assertEqual(sorted(set(result)), sorted(result))

    def test_by_locale_zh(self):
        '''by_locale() for Chinese'''

        ls = self._fake_apt_language_support(['libreoffice-common', 'firefox',
            'ibus'], ['libreoffice-l10n-zh-cn', 'libreoffice-l10n-zh-tw',
                'ibus-sunpinyin', 'ibus-chewing', 'firefox-locale-zh-hans',
                'firefox-locale-zh-hant'])

        result = set(ls.by_locale('zh_CN', True))
        self.assertEqual(result, set(['libreoffice-l10n-zh-cn',
            'ibus-sunpinyin', 'firefox-locale-zh-hans']))

        # accepts both variants of specifying the locale; this is not
        # originally supposed to work, but some programs assume it does
        result = set(ls.by_locale('zh-hans', True))
        self.assertEqual(result, set(['libreoffice-l10n-zh-cn',
            'ibus-sunpinyin', 'firefox-locale-zh-hans']))

        result = set(ls.by_locale('zh_TW', True))
        self.assertEqual(result, set(['libreoffice-l10n-zh-tw',
            'ibus-chewing', 'firefox-locale-zh-hant']))

        result = set(ls.by_locale('zh-hant', True))
        self.assertEqual(result, set(['libreoffice-l10n-zh-tw',
            'ibus-chewing', 'firefox-locale-zh-hant']))

    def test_by_locale_noinstalled(self):
        '''by_locale() without installed packages'''

        ls = self._fake_apt_language_support(['firefox-locale-de',
            'language-pack-de', 'gvfs', 'libreoffice-common', 'firefox'],
            ['language-pack-gnome-de', 'language-pack-kde-de', 'wngerman',
            'wbritish', 'libreoffice-help-de'])
        result = ls.by_locale('de_DE.UTF-8', False)
        self.assertEqual(set(result), set(['language-pack-gnome-de', 'wngerman',
            'libreoffice-help-de']))

    def test_hunspell_de_frami(self):
        '''hunspell-de-frami special case'''

        # if neither is installed, suggest the hunspell-de-de default
        ls = self._fake_apt_language_support(['libreoffice-common'],
                                             ['hunspell-de-de', 'hunspell-de-de-frami'])
        self.assertEqual(ls.missing(), set(['hunspell-de-de']))

        # if the default is installed, it's complete
        ls = self._fake_apt_language_support(['libreoffice-common', 'hunspell-de-de'],
                                             ['hunspell-de-de-frami'])
        self.assertEqual(ls.missing(), set())

        # -frami also suffices
        ls = self._fake_apt_language_support(['libreoffice-common', 'hunspell-de-de-frami'],
                                             ['hunspell-de-de'])
        self.assertEqual(ls.missing(), set())

    def test_by_package(self):
        '''by_package()'''

        # create a fake locales -a
        fake_locale = os.path.join(self.workdir, 'locale')
        with open(fake_locale, 'w') as f:
            f.write('''#!/bin/sh
cat <<EOF
de_DE.UTF-8
pt_BR
zh_CN.GB18030
EOF
''')
        os.chmod(fake_locale, 0o755)
        os.environ['PATH'] = '%s:%s' % (self.workdir, os.getenv('PATH', ''))

        ls = self._fake_apt_language_support([], ['language-pack-de',
            'libreoffice-common', 'libreoffice-help-de',
            'libreoffice-l10n-pt-br', 'libreoffice-l10n-zh-cn',
            'libreoffice-l10n-pt', 'libreoffice-l10n-zh-tw',
            'libreoffice-l10n-uk'])
        result = ls.by_package('libreoffice-common', False)
        self.assertEqual(set(result), set(['libreoffice-help-de',
            'libreoffice-l10n-pt-br', 'libreoffice-l10n-zh-cn',
            'libreoffice-l10n-pt']))

        # no duplicated items in result list
        self.assertEqual(sorted(set(result)), sorted(result))

    def test_all(self):
        '''all()'''

        # create a fake locales -a
        fake_locale = os.path.join(self.workdir, 'locale')
        with open(fake_locale, 'w') as f:
            f.write('''#!/bin/sh
cat <<EOF
de_DE.UTF-8
pt_BR
EOF
''')
        os.chmod(fake_locale, 0o755)
        os.environ['PATH'] = '%s:%s' % (self.workdir, os.getenv('PATH', ''))

        ls = self._fake_apt_language_support(['libreoffice-common', 'firefox',
            'language-pack-de'], 
            ['libreoffice-help-de', 'libreoffice-l10n-pt-br', 'libreoffice-l10n-pt',
            'firefox-locale-de', 'firefox-locale-pt'])
        result = ls.missing(False)
        self.assertEqual(set(result), set(['libreoffice-help-de',
            'libreoffice-l10n-pt-br', 'libreoffice-l10n-pt',
            'firefox-locale-de', 'firefox-locale-pt']))

        # now with already installed packages
        result = ls.missing(True)
        self.assertEqual(set(result), set(['libreoffice-help-de',
            'libreoffice-l10n-pt-br', 'libreoffice-l10n-pt',
            'firefox-locale-de', 'firefox-locale-pt', 'language-pack-de']))

    def test_performance_of_missing(self):
        '''missing() performance for current system'''

        iterations = 20

        r_start = resource.getrusage(resource.RUSAGE_SELF)
        ls = LanguageSupport(self.apt_cache, self.pkg_depends)
        r_init = resource.getrusage(resource.RUSAGE_SELF)
        ls.missing(True)
        r_first = resource.getrusage(resource.RUSAGE_SELF)
        for i in range(iterations):
            ls.missing(True)
        r_iter = resource.getrusage(resource.RUSAGE_SELF)

        init = (r_init.ru_utime + r_init.ru_stime -
            r_start.ru_utime - r_start.ru_stime) * 1000000
        first = (r_first.ru_utime + r_first.ru_stime -
            r_init.ru_utime - r_init.ru_stime) * 1000000
        avg = (r_iter.ru_utime + r_iter.ru_stime -
            r_first.ru_utime - r_first.ru_stime) * 1000000/iterations
        sys.stderr.write('[%iμs init, %iμs first, %iμs avg] ' % (int(init+.5), int(first+.5), int(avg+.5)))

    def test_available_languages(self):
        '''available_languages()'''

        # create a fake locales -a
        fake_locale = os.path.join(self.workdir, 'locale')
        with open(fake_locale, 'w') as f:
            f.write('''#!/bin/sh
cat <<EOF
aa_ER@saaho
de_DE.UTF-8
en_AU.UTF-8
en_US.UTF-8
es_AR.UTF-8
pt_PT.UTF-8
pt_BR
ru_RU.UTF-8
zh_CN.GB18030
zh_TW.UTF-8
EOF
''')
        os.chmod(fake_locale, 0o755)
        os.environ['PATH'] = '%s:%s' % (self.workdir, os.getenv('PATH', ''))

        ls = LanguageSupport(self.apt_cache, self.pkg_depends)
        available = ls.available_languages()
        self.assertEqual(available, set(['aa', 'de', 'en', 'en_AU', 'en_US',
            'es', 'es_AR', 'pt', 'pt_BR', 'pt_PT', 'ru', 'zh_CN', 'zh_TW']))

    def test_countries_for_lang(self):
        '''_countries_for_lang()'''

        self.assertEqual(LanguageSupport._countries_for_lang('yue'), set(['hk']))
        # Serbia has @ variants
        self.assertEqual(LanguageSupport._countries_for_lang('sr'), set(['rs', 'me']))
        self.assertEqual(LanguageSupport._countries_for_lang('de'), 
                set(['at', 'be', 'ch', 'de', 'li', 'lu']))
        en = LanguageSupport._countries_for_lang('en')
        self.assertTrue('us' in en)
        self.assertTrue('gb' in en)
        self.assertTrue('sg' in en)

    def test_available_languages_system(self):
        '''available_languages() for system-installed locales'''

        # we cannot assume much here, just check that it works and does not
        # crash.
        ls = LanguageSupport(self.apt_cache, self.pkg_depends)
        available = ls.available_languages()
        self.assertGreater(len(available), 0)

    def test_apt_cache_add_language_packs(self):
        '''apt_cache_add_language_packs()'''

        # create a fake locales -a
        fake_locale = os.path.join(self.workdir, 'locale')
        with open(fake_locale, 'w') as f:
            f.write('''#!/bin/sh
cat <<EOF
de_DE.UTF-8
es_AR.UTF-8
EOF
''')
        os.chmod(fake_locale, 0o755)
        os.environ['PATH'] = '%s:%s' % (self.workdir, os.getenv('PATH', ''))

        # create a fake apt.Cache
        cache = self._fake_apt_language_support([], 
                ['firefox', 'firefox-locale-de', 'firefox-locale-es', 'firefox-locale-fi',
                'libreoffice-common', 'libreoffice-help-de',
                'language-pack-de', 'language-pack-gnome-de', 'language-pack-kde-de', 
                'gvfs', 'libreoffice-common',
                'wngerman', 'wbritish']).apt_cache

        self.assertEqual(cache.get_changes(), [])

        # if we do not have any changes, it should do nothing
        apt_cache_add_language_packs(None, cache)
        self.assertEqual(cache.get_changes(), [])

        # now install firefox
        c = copy.deepcopy(cache)
        c['firefox'].mark_install()
        apt_cache_add_language_packs(None, c, self.pkg_depends)
        new_pkgs = set([p.name for p in c.get_changes() if p.marked_install])
        # should not install the generic support, just firefox specific stuff
        self.assertEqual(new_pkgs, set(['firefox', 'firefox-locale-de', 'firefox-locale-es']))
        self.assertTrue(c['firefox'].marked_install_from_user)
        self.assertFalse(c['firefox-locale-de'].marked_install_from_user)
        self.assertFalse(c['firefox-locale-es'].marked_install_from_user)

        # install two packages
        c = copy.deepcopy(cache)
        c['gvfs'].mark_install()
        c['libreoffice-common'].mark_install()
        apt_cache_add_language_packs(None, c, self.pkg_depends)
        new_pkgs = set([p.name for p in c.get_changes() if p.marked_install])
        self.assertEqual(new_pkgs, set(['gvfs', 'libreoffice-common',
            'libreoffice-help-de', 'language-pack-gnome-de']))

    def _fake_apt_language_support(self, installed, available):
        '''Return a LanguageSupport object with a fake apt_cache.'''

        class FakeCache(dict):
            def get_changes(self):
                result = []
                for pkg in self.values():
                    if pkg.marked_install:
                        result.append(pkg)
                return result

        class Pkg:
            def __init__(self, name, installed):
                self.name = name
                self.installed = installed
                self.marked_install = False
                self.marked_install_from_user = False

            def __str__(self):
                if self.installed:
                    return '%s (installed)' % self.name
                else:
                    return '%s (available)' % self.name

            def __repr__(self):
                if self.installed:
                    return 'Pkg("%s", True)' % self.name
                else:
                    return 'Pkg("%s", False)' % self.name

            def mark_install(self, from_user=True):
                self.marked_install = True
                self.marked_install_from_user = from_user

        cache = FakeCache()
        for p in available:
            cache[p] = Pkg(p, False)
        for p in installed:
            cache[p] = Pkg(p, True)

        return LanguageSupport(cache, self.pkg_depends)

    def _check_valid_pkgs(self, packages):
        '''Check that all packages in a given list exist.'''

        for pkg in packages:
            self.assertTrue(pkg in self.apt_cache, 'package %s does not exist' % pkg)

unittest.main()



# ImConfig.py (c) 2012-2014 Canonical
# Author: Gunnar Hjalmarsson <gunnarhj@ubuntu.com>
#
# Released under the GPL
#

import os
import subprocess

class ImConfig(object):
    
    def __init__(self):
        pass

    def available(self):
        return os.path.exists('/usr/bin/im-config')

    def getAvailableInputMethods(self):
        inputMethods = subprocess.check_output(['im-config', '-l']).decode().split()
        return sorted(inputMethods)

    def getCurrentInputMethod(self):
        (systemConfig, userConfig, autoConfig) = \
          subprocess.check_output(['im-config', '-m']).decode().split()
        if userConfig != 'missing':
            return userConfig

        """
        no saved user configuration
        let's ask the system and save the system configuration as the user ditto
        """
        system_conf = ''
        if os.path.exists('/usr/bin/fcitx'):
            # Ubuntu Kylin special
            system_conf = 'fcitx'
        elif systemConfig == 'default':
            # Using the autoConfig value might be incorrect if the mode in
            # /etc/default/im-config is 'cjkv'. However, as from im-config 0.24-1ubuntu1
            # the mode is 'auto' for all users of language-selector-gnome.
            system_conf = autoConfig
        elif os.path.exists('/etc/X11/xinit/xinputrc'):
            for line in open('/etc/X11/xinit/xinputrc'):
                if line.startswith('run_im'):
                    system_conf = line.split()[1]
                    break
        if not system_conf:
            system_conf = autoConfig
        self.setInputMethod(system_conf)
        return system_conf

    def setInputMethod(self, im):
        subprocess.call(['im-config', '-n', im])
    
if __name__ == '__main__':
    im = ImConfig()
    print('available input methods: %s' % im.getAvailableInputMethods())
    print('current method: %s' % im.getCurrentInputMethod())
    print("setting method 'fcitx'")
    im.setInputMethod('fcitx')
    print('current method: %s' % im.getCurrentInputMethod())
    print('removing ~/.xinputrc')
    im.setInputMethod('REMOVE')



# (c) 2005 Canonical
# Author: Michael Vogt <michael.vogt@ubuntu.com>
#
# Released under the GPL
#

from __future__ import print_function

import gettext
import grp
import locale
import os
import pwd
import re
import subprocess
import sys
import time
from gettext import gettext as _

from gi.repository import GObject, Gdk, Gtk #, Pango

import apt

import aptdaemon.client
from defer import inline_callbacks
from aptdaemon.enums import *
from aptdaemon.gtk3widgets import AptProgressDialog

from LanguageSelector.LanguageSelector import *
from LanguageSelector.ImConfig import ImConfig
from LanguageSelector.macros import *
from LanguageSelector.utils import language2locale
from LanguageSelector.LangCache import ExceptionPkgCacheBroken
 
(LIST_LANG,                     # language (i18n/human-readable)
 LIST_LANG_INFO                 # the short country code (e.g. de, pt_BR)
 ) = range(2)

(LANGTREEVIEW_LANGUAGE,
 LANGTREEVIEW_CODE) = range(2)
 
(IM_CHOICE,
 IM_NAME) = range(2)


def xor(a,b):
    " helper to simplify the reading "
    return a ^ b

def blockSignals(f):
    " decorator to ensure that the signals are blocked "
    def wrapper(*args, **kwargs):
        args[0]._blockSignals = True
        res = f(*args, **kwargs)
        args[0]._blockSignals = False
        return res
    return wrapper

def honorBlockedSignals(f):
    " decorator to ensure that the signals are blocked "
    def wrapper(*args, **kwargs):
        if args[0]._blockSignals:
            return
        res = f(*args, **kwargs)
        return res
    return wrapper

def insensitive(f):
    """
    decorator to ensure that a given function is run insensitive
    warning: this will not stack well so don't use it for nested
    stuff (a @insensitive func calling a @insensitve one)
    """
    def wrapper(*args, **kwargs):
        args[0].setSensitive(False)
        res = f(*args, **kwargs)
        args[0].setSensitive(True)
        return res
    return wrapper



# intervals of the start up progress
# 3x caching and menu creation
STEPS_UPDATE_CACHE = [33, 66, 100]

class GtkProgress(apt.progress.base.OpProgress):
    def __init__(self, host_window, progressbar, parent,
                 steps=STEPS_UPDATE_CACHE):
        # used for the "one run progressbar"
        self.steps = steps[:]
        self.base = 0
        self.old = 0
        self.next = int(self.steps.pop(0))

        self._parent = parent
        self._window = host_window
        self._progressbar = progressbar
        self._window.realize()
        host_window.get_window().set_functions(Gdk.WMFunction.MOVE)
        self._window.set_transient_for(parent)

    def update(self, percent=None):
        #print(percent)
        #print(self.Op)
        #print(self.SubOp)
        self._window.show()
        self._parent.set_sensitive(False)
        if percent is None:
            self.old = 0
            self._progressbar.pulse()
        else:
            # if the old percent was higher, a new progress was started
            if self.old > percent:
                # set the borders to the next interval
                self.base = self.next
                try:
                    self.next = int(self.steps.pop(0))
                except:
                    pass
            progress = self.base + percent/100 * (self.next - self.base)
            self.old = percent
            self._progressbar.set_fraction(progress/100.0)
        while Gtk.events_pending():
            Gtk.main_iteration()
    def done(self):
        self._parent.set_sensitive(True)
    def hide(self):
        self._window.hide()

class GtkLanguageSelector(LanguageSelectorBase):

    def __init__(self, datadir, options):
        LanguageSelectorBase.__init__(self, datadir)
        self._datadir = datadir

        self.widgets = Gtk.Builder()
        self.widgets.set_translation_domain('language-selector')
        self.widgets.add_from_file(datadir+"/data/LanguageSelector.ui")
        self.widgets.connect_signals(self)

        try:
            in_grp_admin = grp.getgrnam("admin")[2] in os.getgroups()
        except KeyError:
            in_grp_admin = False
        try:
            in_grp_sudo = grp.getgrnam("sudo")[2] in os.getgroups()
        except KeyError:
            in_grp_sudo = False

        self.is_admin = (os.getuid() == 0 or in_grp_sudo or in_grp_admin)

        # see if we have any other human users on this system
        self.has_other_users = False
        num = 0
        for l in pwd.getpwall():
            if l.pw_uid >= 500 and l.pw_uid < 65534:
                num += 1
            if num >= 2:
                self.has_other_users = True
                break
        
        #build the comboboxes (with model)
        combo = self.combobox_locale_chooser
        model = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_STRING)
        cell = Gtk.CellRendererText()
        combo.pack_start(cell, True)
        combo.add_attribute(cell, 'text', LANGTREEVIEW_LANGUAGE)
        combo.set_model(model)
#        self.combo_syslang_dirty = False

#        combo = self.combobox_user_language
#        model = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_STRING)
#        cell = Gtk.CellRendererText()
#        combo.pack_start(cell, True)
#        combo.add_attribute(cell, 'text', COMBO_LANGUAGE)
#        combo.set_model(model)
#        self.combo_userlang_dirty = False
        self.options = options

        # get aptdaemon client
        self.ac = aptdaemon.client.AptClient()

        combo = self.combobox_input_method
        model = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_STRING)
        cell = Gtk.CellRendererText()
        combo.pack_start(cell, True)
        combo.add_attribute(cell, 'text', IM_NAME)
        combo.set_model(model)
        self.ImConfig = ImConfig()
        self._blockSignals = False

        # build the treeview
        self.setupLanguageTreeView()
        self.setupInstallerTreeView()
        self.updateLanguageView()
#        self.updateUserDefaultCombo()
        self.updateLocaleChooserCombo()
        self.check_input_methods()
#        self.updateSyncButton()
        
        # apply button
        self.button_apply.set_sensitive(False)

        # 'Apply System-Wide...' and 'Install/Remove Languages...' buttons
        #
        # Distinction depending on the is_admin test dropped as a
        # temporary(?) fix of https://launchpad.net/bugs/1008344
#        if self.is_admin:
        self.button_apply_system_wide_languages.set_sensitive(True)
        self.button_install_remove_languages.set_sensitive(True)
        self.button_apply_system_wide_locale.set_sensitive(True)
#        else:
#            self.button_apply_system_wide_languages.set_sensitive(False)
#            self.button_install_remove_languages.set_sensitive(False)
#            self.button_apply_system_wide_locale.set_sensitive(False)

        # show it
        self.window_main.show()
        self.setSensitive(False)

        if self.is_admin:
            # check if the package list is up-to-date
            if not self._cache.havePackageLists:
                d = Gtk.MessageDialog(message_type=Gtk.MessageType.INFO,
                                      buttons=Gtk.ButtonsType.CANCEL)
                d.set_modal(True)
                d.set_transient_for(self.window_main)
                d.set_markup("<big><b>%s</b></big>\n\n%s" % (
                    _("No language information available"),
                    _("The system does not have information about the "
                      "available languages yet. Do you want to perform "
                      "a network update to get them now? ")))
                d.set_title=("")
                d.add_button(_("_Update"), Gtk.ResponseType.YES)
                res = d.run()
                d.destroy()
                if res == Gtk.ResponseType.YES:
                    self.setSensitive(False)
                    self.update()
                    self.updateLanguageView()
                    self.setSensitive(True)

            # see if something is missing
            if self.options.verify_installed:
                self.verifyInstalledLangPacks()

        if not self.ImConfig.available():
            self.combobox_input_method.set_sensitive(False)
        self.setSensitive(True)

    def __getattr__(self, name):
        '''Convenient access to GtkBuilder objects'''

        o = self.widgets.get_object(name)
        if o is None:
            raise AttributeError('No such widget: ' + name)
        return o

    def setSensitive(self, value):
        win = self.window_main.get_window()
        if value:
            self.window_main.set_sensitive(True)
            if win:
                win.set_cursor(None)
        else:
            self.window_main.set_sensitive(False)
            if win:
                win.set_cursor(Gdk.Cursor.new(Gdk.CursorType.WATCH))
        while Gtk.events_pending():
            Gtk.main_iteration()

#    @blockSignals
#    def updateSyncButton(self):
#        " check if the sync languages button should be enabled or not "
#        button = self.checkbutton_sync_languages
#        combo = self.combobox_system_language
#        # no admin user, gray out
#        if self.is_admin == False:
#            button.set_active(False)
#            button.set_sensitive(False)
#            combo.set_sensitive(False)
#            return
#        # admin user, check stuff
#        button.set_sensitive(True)
#        combo.set_sensitive(True)
#        # do not enable the keep the same button if the system has other
#        # users or if the language settings are inconsistent already
#        userlang = self.combobox_user_language.get_active()
#        systemlang = self.combobox_system_language.get_active()
#        if (not self.has_other_users and userlang == systemlang):
#            button.set_active(True)
#        else:
#            button.set_active(False)

    def setupInstallerTreeView(self):
        """ do all the treeview setup here """
        def toggle_cell_func(column, cell, model, iter, data):
            langInfo = model.get_value(iter, LIST_LANG_INFO)

            # check for active and inconsitent 
            inconsistent = langInfo.inconsistent
            #if inconsistent:
            #    print("%s is inconsistent" % langInfo.language)

            cell.set_property("active", langInfo.fullInstalled)
            cell.set_property("inconsistent", inconsistent)
            
        def lang_view_func(cell_layout, renderer, model, iter, data):
            langInfo = model.get_value(iter, LIST_LANG_INFO)
            langName = model.get_value(iter, LIST_LANG)
            if (langInfo.changes) :
                markup = "<b>%s</b>" % langName
            else:
                markup = "%s" % langName
            renderer.set_property("markup", markup)

        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(_("Language"), renderer, text=LIST_LANG)
        column.set_property("expand", True)
        column.set_cell_data_func (renderer, lang_view_func, None)
        self.treeview_languages.append_column(column)

        renderer= Gtk.CellRendererToggle()
        renderer.connect("toggled", self.on_toggled)
        column = Gtk.TreeViewColumn(_("Installed"), renderer)
        column.set_cell_data_func (renderer, toggle_cell_func, None)
        self.treeview_languages.append_column(column)
        # build the store
        self._langlist = Gtk.ListStore(str, GObject.TYPE_PYOBJECT)
        self.treeview_languages.set_model(self._langlist)

    def setupLanguageTreeView(self):
        """ do all the treeview setup here """
        def lang_view_func(cell_layout, renderer, model, iter, data):
            langInfo = model.get_value(iter, LANGTREEVIEW_CODE)
            greyFlag = False
            myiter = model.get_iter_first()
            while myiter:
                str = model.get_value(myiter,LANGTREEVIEW_CODE)
                if str == langInfo: 
                    greyFlag = False
                    break
                if str == "en":
                    greyFlag = True
                    break
                myiter = model.iter_next(myiter)
            if greyFlag:
                markup = "<span foreground=\"grey\">%s</span>" \
                       % self._localeinfo.translate(langInfo, native=True, allCountries=True)
            else:
                markup = "%s" % self._localeinfo.translate(langInfo, native=True, allCountries=True)
            renderer.set_property("markup", markup)

        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(_("Language"), renderer, text=LANGTREEVIEW_LANGUAGE)
        column.set_property("expand", True)
        column.set_cell_data_func (renderer, lang_view_func, None)
        self.treeview_locales.append_column(column)

        # build the store
        self._language_options = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_STRING)
        self.treeview_locales.set_model(self._language_options)

    def _get_langinfo_on_cursor(self):
        (path, column) = self.treeview_languages.get_cursor()
        if not path:
            return None
        iter = self._langlist.get_iter(path)
        langInfo = self._langlist.get_value(iter, LIST_LANG_INFO)
        return langInfo

    def debug_pkg_status(self):
        langInfo = self._get_langinfo_on_cursor()
        for pkg in langInfo.languagePkgList.items() :
            print("%s, available: %s, installed: %s, doChange: %s" % (pkg[0], pkg[1].available, pkg[1].installed, pkg[1].doChange))
        print("inconsistent? : %s" % langInfo.inconsistent)

    def check_status(self):
        changed = False
        countInstall = 0
        countRemove = 0
        for (lang, langInfo) in self._langlist:
            if langInfo.changes:
                changed = True
                for item in langInfo.languagePkgList.values():
                    if item.doChange:
                        if item.installed:
                            countRemove = countRemove + 1
                        else:
                            countInstall = countInstall + 1
        #print("%(INSTALL)d to install, %(REMOVE)d to remove" % (countInstall, countRemove))
        # Translators: %(INSTALL)d is parsed; either keep it exactly as is or remove it entirely, but don't translate "INSTALL".
        textInstall = gettext.ngettext("%(INSTALL)d to install", "%(INSTALL)d to install", countInstall) % {'INSTALL': countInstall}
        # Translators: %(REMOVE)d is parsed; either keep it exactly as is or remove it entirely, but don't translate "REMOVE".
        textRemove = gettext.ngettext("%(REMOVE)d to remove", "%(REMOVE)d to remove", countRemove) % {'REMOVE': countRemove}
        if countRemove == 0 and countInstall == 0: 
            self.label_install_remove.set_text("")
        elif countRemove == 0: 
            self.label_install_remove.set_text(textInstall)
        elif countInstall == 0: 
            self.label_install_remove.set_text(textRemove)
        else: 
            # Translators: this string will concatenate the "%n to install" and "%n to remove" strings, you can replace the comma if you need to.
            self.label_install_remove.set_text(_("%s, %s") % (textInstall, textRemove))
        
        if changed:
            self.button_apply.set_sensitive(True)
        else:
            self.button_apply.set_sensitive(False)

#    @honorBlockedSignals
#    @insensitive
#    def on_combobox_system_language_changed(self, widget):
#        #print("on_combobox_system_language_changed()")
#        if self.writeSystemDefaultLang():
#            # queue a restart of gdm (if it is runing) to make the new
#            # locales usable
#            gdmscript = "/etc/init.d/gdm"
#            if os.path.exists("/var/run/gdm.pid") and os.path.exists(gdmscript):
#                self.runAsRoot(["invoke-rc.d","gdm","reload"])
#        self.updateSystemDefaultCombo()
#        if self.checkbutton_sync_languages.get_active() == True:
#            self.combobox_user_language.set_active(self.combobox_system_language.get_active())
#            self.updateUserDefaultCombo()

#    @honorBlockedSignals
#    @insensitive
#    def on_combobox_user_language_changed(self, widget):
#        #print("on_combobox_user_language_changed()")
#        self.check_input_methods()
#        self.writeUserDefaultLang()
#        self.updateUserDefaultCombo()
#        if self.checkbutton_sync_languages.get_active() == True:
#            self.combobox_system_language.set_active(self.combobox_user_language.get_active())
#            self.updateSystemDefaultCombo()

    @blockSignals
    def check_input_methods(self):
        if not self.ImConfig.available():
            return

        combo = self.combobox_input_method
        model = combo.get_model()
        if not model:
            return
        model.clear()

        # find the default
        currentIM = self.ImConfig.getCurrentInputMethod()

        # find out about the other options
        names = dict(xim=_('none'), ibus='IBus', scim='SCIM', hangul='Hangul', thai='Thai')
        for (i, IM) in enumerate(self.ImConfig.getAvailableInputMethods()):
            name = names[IM] if IM in names else IM
            iter = model.append()
            model.set_value(iter, IM_CHOICE, IM)
            model.set_value(iter, IM_NAME, name)
            if IM == currentIM:
                combo.set_active(i)
#        self.check_status()

#    def writeInputMethodConfig(self):
#        """ 
#        write new input method defaults - currently we only support all_ALL
#        """
#        combo = self.combobox_user_language
#        model = combo.get_model()
#        if combo.get_active() < 0:
#            return
#        (lang, code) = model[combo.get_active()]
#        # check if we need to do something
#        new_value = self.checkbutton_enable_input_methods.get_active()
#        if self.imSwitch.enabledForLocale(code) != new_value:
#            if new_value:
#                self.imSwitch.enable(code)
#            else:
#                self.imSwitch.disable(code)
#            #self.showRebootRequired()
#            #self.checkReloginNotification()

#    @honorBlockedSignals
#    def on_checkbutton_enable_input_methods_toggled(self, widget):
#        #print("on_checkbutton_enable_input_methods_toggled()")
#        active = self.checkbutton_enable_input_methods.get_active()
#        self.combo_userlang_dirty = True
#        self.setSensitive(False)
#        self.writeInputMethodConfig()
#        self.setSensitive(True)

#    @honorBlockedSignals
#    def on_checkbutton_sync_languages_toggled(self, widget):
#        #print("on_checkbutton_sync_languages_toggled()")
#        if self.checkbutton_sync_languages.get_active() == True:
#            self.combobox_user_language.set_active(self.combobox_system_language.get_active())
#            self.updateSystemDefaultCombo()
        
    def build_commit_lists(self):
        print(self._cache.get_changes())

        try:
            for (lang, langInfo) in self._langlist:
                self._cache.tryChangeDetails(langInfo)
        except ExceptionPkgCacheBroken:
            self.error(
                _("Software database is broken"),
                _("It is impossible to install or remove any software. "
                  "Please use the package manager \"Synaptic\" or run "
                  "\"sudo apt-get install -f\" in a terminal to fix "
                  "this issue at first."))
            sys.exit(1)
        (to_inst, to_rm) = self._cache.getChangesList()
        #print("inst: %s" % to_inst)
        #print("rm: %s" % to_rm)
        print(self._cache.get_changes())
        return (to_inst, to_rm)

    def error(self, summary, msg):
        d = Gtk.MessageDialog(message_type=Gtk.MessageType.ERROR,
                              buttons=Gtk.ButtonsType.CLOSE)
        d.set_modal(True)
        d.set_transient_for(self.window_main)
        d.set_markup("<big><b>%s</b></big>\n\n%s" % (summary, msg))
        d.set_title=("")
        d.run()
        d.destroy()

    def _show_error_dialog(self, error):
        msg = str(error)
        self.error(msg, "")

    def verify_commit_lists(self, inst_list, rm_list):
        """ verify if the selected package can actually be installed """
        res = True
        try:
            for pkg in inst_list:
                if pkg in self._cache:
                    self._cache[pkg].mark_install()
            for pkg in rm_list:
                if pkg in self._cache:
                    self._cache[pkg].mark_delete()
        except SystemError:
            res = False

        # undo the selections
        self._cache.clear()
        if self._cache._depcache.broken_count != 0:
            self.error(_("Could not install the selected language support"),
                       _("This is perhaps a bug of this application. Please "
                         "file a bug report at "
                         "https://bugs.launchpad.net/ubuntu/+source/language-selector/+filebug"))

            # something went pretty bad, re-get a cache
            progress = GtkProgress(self.dialog_progress,
                                   self.progressbar_cache,
                                   self.window_main)
            self._cache = apt.Cache(self._localeinfo, progress)
            progress.hide()
            res = False
        return res

    def commitAllChanges(self):
        """ 
        commit helper, builds the commit lists, verifies it
        
        returns the number of install/removed packages
        """
        self.setSensitive(False)
        # install the new packages (if any)
        (inst_list, rm_list) = self.build_commit_lists()
        if not self.verify_commit_lists(inst_list, rm_list):
            self.error(
                _("Could not install the full language support"),
                _("Usually this is related to an error in your "
                  "software archive or software manager. Check your "
                  "preferences in Software Sources (click the icon "
                  "at the very right of the top bar and select "
                  "\"System Settings... -> Software Sources\")."))
            self.setSensitive(True)
            return 0
        #print("inst_list: %s " % inst_list)
        #print("rm_list: %s " % rm_list)
        self.commit(inst_list, rm_list)

        self.setSensitive(True)
        return len(inst_list)+len(rm_list)

    
    def _run_transaction(self, transaction):
        dia = AptProgressDialog(transaction)
        dia.set_transient_for(self.window_main)
        dia.connect("finished", self._on_finished)
        dia.run(error_handler=self._on_error)
        
    def _wait_for_aptdaemon_finish(self):
        while not self._transaction_finished:
            while Gtk.events_pending():
                Gtk.main_iteration()
            time.sleep(0.02)

    def _on_finished(self, dialog):
        dialog.hide()
        self._transaction_finished = True

    def _on_error(self, error):
        if hasattr(error, 'get_dbus_name') and error.get_dbus_name() == \
                    'org.freedesktop.PolicyKit.Error.NotAuthorized':
                self.error(
                        _("Could not install the full language support"),
                        _('Failed to authorize to install packages.'))
        else:
            self.error(
                    _("Could not install the full language support"),
                    str(error))
        self._transaction_finished = True

    def update_aptdaemon(self):
        self._transaction_finished = False
        self._update_aptdaemon()
        self._wait_for_aptdaemon_finish()

    @inline_callbacks
    def _update_aptdaemon(self):
        try:
            trans = yield self.ac.update_cache(defer=True)
            self._run_transaction(trans)
        except Exception as e:
            self._show_error_dialog(e)

    def commit_aptdaemon(self, inst, rm):
        self._transaction_finished = False
        self._commit_aptdaemon(inst, rm)
        self._wait_for_aptdaemon_finish()

    @inline_callbacks
    def _commit_aptdaemon(self, inst, rm):
        if len(inst) == 0 and len(rm) == 0:
            return
        try:
            trans = yield self.ac.commit_packages(
                install=inst, reinstall=[], remove=rm, purge=[], upgrade=[],
                downgrade=[], defer=True)
            self._run_transaction(trans)
        except Exception as e:
            self._show_error_dialog(e)

    # we default with update/commit to aptdaemon
    update = update_aptdaemon
    commit = commit_aptdaemon

    def hide_on_delete(self, widget, event):
        return Gtk.Widget.hide_on_delete(widget)

    def verifyInstalledLangPacks(self):
        """ called at the start to inform about possible missing
            langpacks (e.g. gnome/kde langpack transition)
        """
        #print("verifyInstalledLangPacks")
        missing = self.getMissingLangPacks()

        #print("Missing: %s " % missing)
        if len(missing) > 0:
            # FIXME: add "details"
            d = Gtk.MessageDialog(message_type=Gtk.MessageType.QUESTION)
            d.set_modal(True)
            d.set_transient_for(self.window_main)
            d.set_markup("<big><b>%s</b></big>\n\n%s" % (
                _("The language support is not installed completely"),
                _("Some translations or writing aids available for your "
                  "chosen languages are not installed yet. Do you want "
                  "to install them now?")))
            d.add_buttons(_("_Remind Me Later"), Gtk.ResponseType.NO,
                          _("_Install"), Gtk.ResponseType.YES)
            d.set_default_response(Gtk.ResponseType.YES)
            d.set_title("")
            expander = Gtk.Expander.new(_("Details"))
            scroll = Gtk.ScrolledWindow()
            scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
            scroll.set_min_content_height(130)
            textview = Gtk.TextView()
            textview.set_cursor_visible(False)
            textview.set_editable(False)
            buf = textview.get_buffer()
            pkgs = ""
            for pkg in missing:
                pkgs += "%s\n" % pkg
            buf.set_text(pkgs, -1)
            buf.place_cursor(buf.get_start_iter())
            expander.add(scroll)
            scroll.add(textview)
            d.get_message_area().pack_start(expander, True, True, 0)
            expander.show_all()
            res = d.run()
            d.destroy()
            if res == Gtk.ResponseType.YES:
                self.setSensitive(False)
                self.commit(missing, [])
                self.updateLanguageView()
                self.setSensitive(True)

    def updateLanguageView(self):
        #print("updateLanguageView()")
        self._langlist.clear()

        progress = GtkProgress(self.dialog_progress, self.progressbar_cache,
                               self.window_main)
        try:
            if self.is_admin:
                self.openCache(progress)
                progress.hide()
            else:
                self.openCache(None)
        except ExceptionPkgCacheBroken:
            self.error(
                _("Software database is broken"),
                _("It is impossible to install or remove any software. "
                  "Please use the package manager \"Synaptic\" or run "
                  "\"sudo apt-get install -f\" in a terminal to fix "
                  "this issue at first."))
            sys.exit(1)

        languageList = self._cache.getLanguageInformation()
        #print("ll size: ", len(languageList))
        #print("ll type: ", type(languageList))
        for lang in languageList:
            #print("langInfo: %s (%s)" % (lang.language, lang.languageCode))
            #inconsistent = lang.inconsistent
            #if inconsistent:
            #    print("inconsistent", lang.language)

            # hack for Vietnamese users; see https://launchpad.net/bugs/783090
            # Even if calling self._localeinfo.translate() with native=True triggers
            # a UnicodeWarning for Norwegian Bokmaal, the output should be correct.
            lang_name = None
            if 'LANGUAGE' in os.environ:
                current_language = os.environ['LANGUAGE']
                if re.match('vi[^a-z]', current_language):
                    lang_name = self._localeinfo.translate(lang.languageCode, native=True)
            if not lang_name:
                lang_name = self._localeinfo.translate(lang.languageCode)

            self._langlist.append([lang_name, lang])
        self._langlist.set_sort_column_id(LIST_LANG, Gtk.SortType.ASCENDING)

    def writeSystemFormats(self):
        combo = self.combobox_locale_chooser
        model = combo.get_model()
        if combo.get_active() < 0:
            return False
        (lang, code) = model[combo.get_active()]
        old_code = self._localeinfo.getSystemDefaultLanguage()[0]
        # no changes, nothing to do
        macr = macros.LangpackMacros(self._datadir, old_code)
        if macr["LOCALE"] == code:
            return False
        self.writeSysFormatsSetting(sysFormats=code)
        return True

    def writeUserFormats(self):
        combo = self.combobox_locale_chooser
        model = combo.get_model()
        if combo.get_active() < 0:
            return
        (lang, code) = model[combo.get_active()]
        temp = self._localeinfo.getUserDefaultLanguage()[0]
        if temp == None:
            old_code = self._localeinfo.getSystemDefaultLanguage()[0]
        else:
            old_code = temp
        # no changes, nothing to do
        macr = macros.LangpackMacros(self._datadir, old_code)
        if macr["LOCALE"] == code:
            return False
        self.writeUserFormatsSetting(userFormats=code)
        return True

    def writeSystemLanguage(self, languageString):
        (formats_locale, old_string) = self._localeinfo.getSystemDefaultLanguage()
        # no changes, nothing to do
        if old_string == languageString:
            return False
        self.writeSysLanguageSetting(sysLanguage=languageString)
        if self._localeinfo.isSetSystemFormats():
            return True
        # set the system formats (certain LC_* variables) explicitly
        # in order to prevent surprises when LANG is changed
        self.writeSysFormatsSetting(sysFormats=formats_locale)
        return True

    def writeUserLanguage(self, languageString):
        temp = self._localeinfo.getUserDefaultLanguage()[1]
        if len(temp) == 0:
            old_string = self._localeinfo.getSystemDefaultLanguage()[1]
        else:
            old_string = temp
        # no changes, nothing to do
        if old_string == languageString:
            return False
        self.writeUserLanguageSetting(userLanguage=languageString)
        return True

    @blockSignals
    def updateLocaleChooserCombo(self):
        #print("updateLocaleChooserCombo()")
        combo = self.combobox_locale_chooser
        #XXX get_cell_renderers does not exist in GTK3
        #cell = combo.get_child().get_cell_renderers()[0]
        # FIXME: use something else than a hardcoded value here
        #cell.set_property("wrap-width",300)
        #cell.set_property("wrap-mode",Pango.WRAP_WORD)
        model = combo.get_model()

        if not model:
            return

        model.clear()

        # find the default
        defaultLangName = None
        (defaultLangCode, languageString) = self._localeinfo.getUserDefaultLanguage()
        if len(defaultLangCode) == 0:
            defaultLangCode = self._localeinfo.getSystemDefaultLanguage()[0]
        if len(defaultLangCode) > 0:
            macr = macros.LangpackMacros(self._datadir, defaultLangCode)
            defaultLangCode = macr["LOCALE"]
            defaultLangName = self._localeinfo.translate(defaultLangCode, native=True)

        # find out about the other options        

        """ languages for message translation """
        self._language_options.clear()
        options = subprocess.check_output(
            ['/usr/share/language-tools/language-options'],
            universal_newlines=True)
        mylist = []
        for (i, option) in enumerate( options.split("\n") ):
            mylist.append([self._localeinfo.translate(option, native=True), option])
        if len(languageString) > 0:
            self.userEnvLanguage = languageString
            languages = languageString.split(":")
        else:
            if 'LANGUAGE' in os.environ:
                self.userEnvLanguage = os.environ.get("LANGUAGE")
                languages = self.userEnvLanguage.split(":")
            else:
                self.userEnvLanguage = self._localeinfo.makeEnvString(defaultLangCode)
                languages = self.userEnvLanguage.split(":")
        mylist_sorted = self.bubbleSort(mylist, languages)
        for i in mylist_sorted:
            self._language_options.append(i)

        """ locales for misc. format preferences """
        for (i, loc) in enumerate( self._localeinfo.generated_locales() ):
            iter = model.append()
            model.set_value(iter, LANGTREEVIEW_LANGUAGE,
                    self._localeinfo.translate(loc, native=True))
            model.set_value(iter, LANGTREEVIEW_CODE, loc)
            if (defaultLangName and
                   self._localeinfo.translate(loc, native=True) == defaultLangName):
                combo.set_active(i)
        self.updateExampleBox()

    def bubbleSort(self, sortlist, presort=None):
        """
        Sort the list 'sortlist' using bubble sort.
        Optionally, if a list 'presort' is given, put this list first and bubble sort the rest.
        """
        for i in range(0,len(sortlist)-1):
            for j in range(0,len(sortlist)-1):
                data1 = sortlist[j][1]  
                data2 = sortlist[j+1][1]
                try:
                    v1 = presort.index(data1)
                except:
                    v1 = 100000
                try:
                    v2 = presort.index(data2)
                except:
                    v2 = 100000
                if (v1>v2):
                    sortlist[j],sortlist[j+1] = sortlist[j+1], sortlist[j]   
                elif (v1 >= 100000 and v2 >= 100000 and data1 > data2):
                    sortlist[j],sortlist[j+1] = sortlist[j+1], sortlist[j]
        return sortlist

#        # reset the state of the apply button
#        self.combo_syslang_dirty = False
#        self.check_status()

    # FIXME: updateUserDefaultCombo and updateSystemDefaultCombo
    #        duplicate too much code
#    @blockSignals
#    def updateUserDefaultCombo(self):
#        #print("updateUserDefault()")
#        combo = self.combobox_user_language
#        cell = combo.get_child().get_cell_renderers()[0]
#        # FIXME: use something else than a hardcoded value here
#        cell.set_property("wrap-width",300)
#        cell.set_property("wrap-mode",Pango.WRAP_WORD)
#        model = combo.get_model()
#        model.clear()

#        # find the default
#        defaultLangName = None
#        defaultLangCode = self.getUserDefaultLanguage()
#        if defaultLangCode == None:
#            defaultLangCode = self.getSystemDefaultLanguage()
#        if defaultLangCode:
#            defaultLangName = self._localeinfo.translate(defaultLangCode)

#        # find out about the other options        
#        for (i, loc) in enumerate(self._localeinfo.generated_locales()):
#            iter = model.append()
#            model.set(iter,
#                      COMBO_LANGUAGE,self._localeinfo.translate(loc),
#                      COMBO_CODE, loc)
#            if (defaultLangName and 
#                   self._localeinfo.translate(loc) == defaultLangName):
#                combo.set_active(i)
#            
#        # reset the state of the apply button
#        self.combo_userlang_dirty = False
#        self.check_status()

    def updateExampleBox(self):
        combo = self.combobox_locale_chooser
        model = combo.get_model()
        if combo.get_active() < 0:
            return
        self.label_example_currency.set_text('')
        self.label_example_number.set_text('')
        self.label_example_date.set_text('')
        (lang, code) = model[combo.get_active()]
        macr = macros.LangpackMacros(self._datadir, code)
        mylocale = macr["SYSLOCALE"]
        try:
            locale.setlocale(locale.LC_ALL, mylocale)
        except locale.Error:
            self.label_example_date.set_text('[ '
            + _("Failed to apply the '%s' format\n"
                "choice. The examples may show up if you\n"
                "close and re-open Language Support.") % mylocale + ' ]')
            return
        self.label_example_currency.set_text(locale.currency(20457.99, grouping=True))
        self.label_example_number.set_text(locale.format("%.2f", 1234567.89, grouping=True))
        self.label_example_date.set_text(time.strftime(locale.nl_langinfo(locale.D_T_FMT)))


    ####################################################
    # window_installer signal handlers                 #
    ####################################################

    def on_treeview_languages_row_activated(self, treeview, path, view_column):
        self.on_toggled(None, path.to_string())

    # the global toggle
    def on_toggled(self, renderer, path_string):
        """ called when on install toggle """
        iter = self._langlist.get_iter_from_string(path_string)
        langInfo = self._langlist.get_value(iter, LIST_LANG_INFO)

        # special handling for inconsistent state
        if langInfo.inconsistent :
            for pkg in langInfo.languagePkgList.values() :
                if (pkg.available and not pkg.installed) : 
                    pkg.doChange = True
        elif langInfo.fullInstalled :
            for pkg in langInfo.languagePkgList.values() :
                if (pkg.available) :
                    if (not pkg.installed and pkg.doChange) :
                        pkg.doChange = False
                    elif (pkg.installed and not pkg.doChange) :
                        pkg.doChange = True
        else :
            for pkg in langInfo.languagePkgList.values() :
                if (pkg.available) :
                    if (pkg.installed and pkg.doChange) :
                        pkg.doChange = False
                    elif (not pkg.installed and not pkg.doChange) :
                        pkg.doChange = True

        self.check_status()
        self.treeview_languages.queue_draw()
        #self.debug_pkg_status()

    def on_button_cancel_clicked(self, widget):
        #print("button_cancel")
        self.window_installer.hide()

    def on_button_apply_clicked(self, widget):
        self.window_installer.hide()
        if self.commitAllChanges() > 0:
            self.updateLanguageView()
        self.updateLocaleChooserCombo()
        #self.updateSystemDefaultCombo()

    ####################################################
    # window_main signal handlers                      #
    ####################################################
    def on_delete_event(self, event, data):
        app = self.window_main.get_application()
        if app:
            app.remove_window(self.window_main)

    def on_button_quit_clicked(self, widget):
        app = self.window_main.get_application()
        if app:
            app.remove_window(self.window_main)

    @honorBlockedSignals
    def on_window_main_key_press_event(self, widget, event):
        keyname = Gdk.keyval_name(event.keyval)
        if event.get_state() & Gdk.ModifierType.CONTROL_MASK and keyname == "w":
            app = self.window_main.get_application()
            if app:
                app.remove_window(self.window_main)
        if (event.get_state() | Gdk.ModifierType.MOD2_MASK) == Gdk.ModifierType.MOD2_MASK and keyname == "Escape":
            app = self.window_main.get_application()
            if app:
                app.remove_window(self.window_main)
        return None

    ####################################################
    # window_main signal handlers (Language tab)       #
    ####################################################
#    @honorBlockedSignals
#    def on_treeview_locales_drag_failed(self, widget):
#        return None

#    @honorBlockedSignals
#    def on_treeview_locales_drag_begin(self, widget):
#        return None

#    @honorBlockedSignals
#    def on_treeview_locales_drag_drop(self, widget):
#        return None

    @honorBlockedSignals
    def on_treeview_locales_drag_end(self, widget, drag_content):
        #print("on_treeview_locales_drag_end")
        model = widget.get_model()
        myiter = model.get_iter_first()
        envLanguage = ""
        while myiter:   
            str = model.get_value(myiter,LANGTREEVIEW_CODE)
            if (envLanguage != ""):
                envLanguage = envLanguage + ":"
            envLanguage = envLanguage + str
            if str == "en":
                break
            myiter = model.iter_next(myiter)
        #print(envLanguage)
        self.writeUserLanguage(envLanguage)
        self.userEnvLanguage = envLanguage
        self.check_input_methods()
        self.updateLocaleChooserCombo()
        #os.environ["LANGUAGE"]=envLanguage

#    @honorBlockedSignals
#    def on_treeview_locales_drag_leave(self, widget):
#        return None

#    @honorBlockedSignals
#    def on_treeview_locales_drag_data_received(self, widget):
#        return None

    @honorBlockedSignals
    @insensitive
    def on_button_apply_system_wide_languages_clicked(self, widget):
        self.writeSystemLanguage(self.userEnvLanguage)
        return None

    def on_button_install_remove_languages_clicked(self, widget):
        self.window_installer.show()

    @honorBlockedSignals
    def on_combobox_input_method_changed(self, widget):
        combo = self.combobox_input_method
        model = combo.get_model()
        if combo.get_active() < 0:
            return
        (IM_choice, IM_name) = model[combo.get_active()]
        self.ImConfig.setInputMethod(IM_choice)

    
    ####################################################
    # window_main signal handlers (Text tab)           #
    ####################################################
    @honorBlockedSignals
    @insensitive
    def on_combobox_locale_chooser_changed(self, widget):
        self.check_input_methods()
        self.writeUserFormats()
        self.updateLocaleChooserCombo()
        self.updateExampleBox()

    @honorBlockedSignals
    @insensitive
    def on_button_apply_system_wide_locale_clicked(self, widget):
        self.writeSystemFormats()
        return None




# (c) 2006 Canonical
# Author: Michael Vogt <michael.vogt@ubuntu.com>
#
# Released under the GPL
#

from __future__ import print_function
from __future__ import absolute_import

import warnings
warnings.filterwarnings("ignore", "apt API not stable yet", FutureWarning)
import apt
import dbus
import os

import language_support_pkgs
from LanguageSelector.LocaleInfo import LocaleInfo
import LanguageSelector.LangCache
from LanguageSelector.utils import *
from LanguageSelector import macros

# the language-selector abstraction
class LanguageSelectorBase(object):
    """ base class for language-selector code """

    def __init__(self, datadir=""):
        self._datadir = datadir
        # load the localeinfo "database"
        self._localeinfo = LocaleInfo("languagelist", self._datadir)
        self._cache = None

    def openCache(self, progress):
        self._cache = LanguageSelector.LangCache.LanguageSelectorPkgCache(self._localeinfo, progress)

    def getMissingLangPacks(self):
        """
        return a list of language packs that are not installed
        but should be installed
        """
        if self._datadir:
            ls = language_support_pkgs.LanguageSupport(self._cache, 
                    os.path.join(self._datadir, "data", "pkg_depends"))
        else:
            ls = language_support_pkgs.LanguageSupport(self._cache)
        missing = []
        for pack in ls.missing():  # ls.missing() returns a set; we need a list
            missing.append(pack)
        return missing

    def writeSysFormatsSetting(self, sysFormats):
        """ write various LC_* variables (e.g. de_DE.UTF-8) """
        bus = dbus.SystemBus()
        obj = bus.get_object('com.ubuntu.LanguageSelector','/')
        iface = dbus.Interface(obj,dbus_interface="com.ubuntu.LanguageSelector")
        iface.SetSystemDefaultFormatsEnv(sysFormats)

    def writeSysLanguageSetting(self, sysLanguage):
        """ write the system "LANGUAGE" and "LANG" variables """
        bus = dbus.SystemBus()
        obj = bus.get_object('com.ubuntu.LanguageSelector','/')
        iface = dbus.Interface(obj,dbus_interface="com.ubuntu.LanguageSelector")
        iface.SetSystemDefaultLanguageEnv(sysLanguage)

    def writeUserFormatsSetting(self, userFormats):
        """ write various LC_* variables (e.g. de_DE.UTF-8) """
        uid = os.getuid()
        if uid == 0:
            warnings.warn("No formats locale saved for user '%s'." % os.getenv('USER'))
            return
        bus = dbus.SystemBus()
        obj = bus.get_object('org.freedesktop.Accounts',
                            '/org/freedesktop/Accounts/User%i' % uid)
        iface = dbus.Interface(obj, dbus_interface='org.freedesktop.Accounts.User')
        macr = macros.LangpackMacros(self._datadir, userFormats)
        iface.SetFormatsLocale(macr['SYSLOCALE'])

    def writeUserLanguageSetting(self, userLanguage):
        """ write the user "LANGUAGE" and "LANG" variables """
        uid = os.getuid()
        if uid == 0:
            warnings.warn("No language saved for user '%s'." % os.getenv('USER'))
            return
        bus = dbus.SystemBus()
        obj = bus.get_object('org.freedesktop.Accounts',
                            '/org/freedesktop/Accounts/User%i' % uid)
        iface = dbus.Interface(obj, dbus_interface='org.freedesktop.Accounts.User')
        iface.SetLanguage(self.validateLangList(userLanguage))

    def validateLangList(self, userLanguage):
        """
        remove possible non-English elements after the first
        English element
        """
        tmp = []
        is_eng = False
        for lang in userLanguage.split(':'):
            if lang.startswith('en_') or lang == 'en':
                tmp.append(lang)
                is_eng = True
            elif not is_eng:
                tmp.append(lang)
        validatedLangList = ':'.join(tmp)
        if validatedLangList != userLanguage:
            warnings.warn('The language list was modified by the program, since there '
                        + 'should not be any non-English items after an English item.')
        return validatedLangList


if __name__ == "__main__":
    lsb = LanguageSelectorBase(datadir="..")
    lsb.openCache(apt.progress.OpProgress())
    print(lsb.verifyPackageLists())





'''macros.py: Generate macro values from configuration values and provide
substitution functions.

The following macros are available:

  LCODE CCODE PKGCODE LOCALE
'''

from __future__ import print_function

import os
import re

def _file_map(file, key, sep = None):
    '''Look up key in given file ("key value" lines). Throw an exception if
    key was not found.'''

    val = None
    for l in open(file):
        try:
            (k, v) = l.split(sep)
        except ValueError:
            continue
        # sort out comments
        if k.find('#') >= 0 or v.find('#') >= 0:
            continue
        if k == key:
            val = v.strip()
    if val == None:
        raise KeyError('Key %s not found in %s' % (key, file))
    return val

class LangcodeMacros:
    
    LANGCODE_TO_LOCALE = '/usr/share/language-selector/data/langcode2locale'

    def __init__(self, langCode):
        self.macros = {}
        locales = {}
        for l in open(self.LANGCODE_TO_LOCALE):
            try:
                l = l.rstrip()
                (k, v) = l.split(':')
            except ValueError:
                continue
            if k.find('#') >= 0 or v.find('#') >= 0:
                continue
            if not k in locales:
                locales[k] = []
            locales[k].append(v)
        self['LOCALES'] = locales[langCode]

    def __getitem__(self, item):
        # return empty string as default
        return self.macros.get(item, '')

    def __setitem__(self, item, value):
        self.macros[item] = value

    def __contains__(self, item):
        return self.macros.__contains__(item)

class LangpackMacros:
    def __init__(self, datadir,locale):
        '''Initialize values of macros.

        This uses information from maps/, config/, some hardcoded aggregate
        strings (such as package names), and some external input:
        
        - locale: Standard locale representation (e. g. pt_BR.UTF-8)
                  Format is: ll[_CC][.UTF-8][@variant]
        '''

        self.LOCALE_TO_LANGPACK = os.path.join(datadir, 'data', 'locale2langpack')
        self.macros = {}
        self['LCODE'] = ''      # the language code
        self['CCODE'] = ''      # the country code if present
        self['VARIANT'] = ''    # the part behind the @ if present
        self['LOCALE'] = ''     # the locale with the .UTF-8 stripped off
        self['PKGCODE'] = ''    # the language code used in the language-packs
        self['SYSLOCALE'] = ''  # a generated full locale identifier, e.g. ca_ES.UTF-8@valencia
        # 'C' and 'POSIX' are not supported as locales, fall back to 'en_US'
        if locale == 'C' or locale == 'POSIX':
            locale = 'en_US'
        if '@' in locale:
            (locale, self['VARIANT']) = locale.split('@')
        if '.' in locale:
            locale = locale.split('.')[0]
        if '_' in locale:
            (self['LCODE'], self['CCODE']) = locale.split('_')
        else:
            self['LCODE'] = locale
        if len(self['VARIANT']) > 0:
            self['LOCALE'] = "%s@%s" % (locale, self['VARIANT'])
        else:
            self['LOCALE'] = locale
        # generate a SYSLOCALE from given components
        if len(self['LCODE']) > 0:
            if len(self['CCODE']) > 0:
                self['SYSLOCALE'] = "%s_%s.UTF-8" % (self["LCODE"], self["CCODE"])
            else:
                self['SYSLOCALE'] = "%s.UTF-8" % self['LCODE']
            if len(self['VARIANT']) > 0:
                self['SYSLOCALE'] = "%s@%s" % (self['SYSLOCALE'], self['VARIANT'])

        # package code
        try:
            self['PKGCODE'] = _file_map(self.LOCALE_TO_LANGPACK, self['LOCALE'], ':')
        except KeyError:
            self['PKGCODE'] = self['LCODE']

    def __getitem__(self, item):
        # return empty string as default
        return self.macros.get(item, '')

    def __setitem__(self, item, value):
        self.macros[item] = value

    def __contains__(self, item):
        return self.macros.__contains__(item)

    def subst_string(self, s):
        '''Substitute all macros in given string.'''

        re_macro = re.compile('%([A-Z]+)%')
        while 1:
            m = re_macro.search(s)
            if m:
                s = s[:m.start(1)-1] + self[m.group(1)] + s[m.end(1)+1:]
            else:
                break

        return s

    def subst_file(self, file):
        '''Substitute all macros in given file.'''

        s = open(file).read()
        open(file, 'w').write(self.subst_string(s))

    def subst_tree(self, root):
        '''Substitute all macros in given directory tree.'''

        for path, dirs, files in os.walk(root):
            for f in files:
                self.subst_file(os.path.join(root, path, f))

if __name__ == '__main__':
    datadir = '/usr/share/language-selector'
    for locale in ['de', 'de_DE', 'de_DE.UTF-8', 'de_DE.UTF-8@euro', 'fr_BE@latin', 'zh_CN.UTF-8', 'zh_TW.UTF-8', 'zh_HK.UTF-8', 'invalid_Locale']:
        l = LangpackMacros(datadir, locale)
        print('-------', locale, '---------------')
        template = '"%PKGCODE%: %LCODE% %CCODE% %VARIANT% %LOCALE% %SYSLOCALE%"'
        print('string:', l.subst_string(template))

        open('testtest', 'w').write(template)
        l.subst_file('testtest')
        print('file  :', open('testtest').read())
        os.unlink('testtest')




# LocaleInfo.py (c) 2006 Canonical, released under the GPL
#
# a helper class to get locale info

from __future__ import print_function
from __future__ import absolute_import

import re            
import subprocess
import gettext
import os
import pwd
import sys
import dbus
import warnings

from LanguageSelector import macros

from gettext import gettext as _
from xml.etree.ElementTree import ElementTree

class LocaleInfo(object):
    " class with handy functions to parse the locale information "
    
    environments = ["/etc/default/locale"]
    def __init__(self, languagelist_file, datadir):
        self._datadir = datadir
        LANGUAGELIST = os.path.join(datadir, 'data', languagelist_file)
        # map language to human readable name, e.g.:
        # "pt"->"Portuguise", "de"->"German", "en"->"English"
        self._lang = {}

        # map country to human readable name, e.g.:
        # "BR"->"Brasil", "DE"->"Germany", "US"->"United States"
        self._country = {}
        
        # map locale (language+country) to the LANGUAGE environment, e.g.:
        # "pt_PT"->"pt_PT:pt:pt_BR:en_GB:en"
        self._languagelist = {}
        
        # read lang file
        et = ElementTree(file="/usr/share/xml/iso-codes/iso_639_3.xml")
        it = et.iter('iso_639_3_entry')
        for elm in it:
            if "common_name" in elm.attrib:
                lang = elm.attrib["common_name"]
            else:
                lang = elm.attrib["name"]
            if "part1_code" in elm.attrib:
                code = elm.attrib["part1_code"]
            else:
                code = elm.attrib["id"]
            self._lang[code] = lang
        # Hack for Chinese langpack split
        # Translators: please translate 'Chinese (simplified)' and 'Chinese (traditional)' so that they appear next to each other when sorted alphabetically.
        self._lang['zh-hans'] = _("Chinese (simplified)")
        # Translators: please translate 'Chinese (simplified)' and 'Chinese (traditional)' so that they appear next to each other when sorted alphabetically.
        self._lang['zh-hant'] = _("Chinese (traditional)")
        # end hack
        
        # read countries
        et = ElementTree(file="/usr/share/xml/iso-codes/iso_3166.xml")
        it = et.iter('iso_3166_entry')
        for elm in it:
            if "common_name" in elm.attrib:
                descr = elm.attrib["common_name"]
            else:
                descr = elm.attrib["name"]
            if "alpha_2_code" in elm.attrib:
                code = elm.attrib["alpha_2_code"]
            else:
                code = elm.attrib["alpha_3_code"]
            self._country[code] = descr
            
        # read the languagelist
        with open(LANGUAGELIST) as f:
            for line in f:
                tmp = line.strip()
                if tmp.startswith("#") or tmp == "":
                    continue
                w = tmp.split(";")
                # FIXME: the latest localechoosers "languagelist" does
                # no longer have this field for most languages, so
                # deal with it and don't set LANGUAGE then
                # - the interessting question is what to do
                # if LANGUAGE is already set and the new
                localeenv = w[6].split(":")
                #print(localeenv)
                self._languagelist[localeenv[0]] = '%s' % w[6]

    def lang(self, code):
        """ map language code to language name """
        if code in self._lang:
            return self._lang[code]
        return ""

    def country(self, code):
        """ map country code to country name"""
        if code in self._country:
            return self._country[code]
        return ""

    def generated_locales(self):
        """ return a list of locales available on the system
            (running locale -a) """
        locales = []
        p = subprocess.Popen(["locale", "-a"], stdout=subprocess.PIPE,
                             universal_newlines=True)
        for line in p.communicate()[0].split("\n"):
            tmp = line.strip()
            if tmp.find('.utf8') < 0:
                continue
            # we are only interessted in the locale, not the codec
            macr = macros.LangpackMacros(self._datadir, tmp)
            locale = macr["LOCALE"]
            if not locale in locales:
                locales.append(locale)
        #print(locales)
        return locales

    def translate_language(self, lang):
        "return translated language"
        if lang in self._lang:
            lang_name = gettext.dgettext('iso_639', self._lang[lang])
            if lang_name == self._lang[lang]:
                lang_name = gettext.dgettext('iso_639_3', self._lang[lang])
            return lang_name
        else:
            return lang

    def translate_country(self, country):
        """
        return translated language and country of the given
        locale into the given locale, e.g. 
        (Deutsch, Deutschland) for de_DE
        """

#        macr = macros.LangpackMacros(self._datadir, locale)

#        #(lang, country) = locale.split("_")
#        country = macr['CCODE']
#        current_language = None
#        if "LANGUAGE" in os.environ:
#            current_language = os.environ["LANGUAGE"]
#        os.environ["LANGUAGE"]=locale
#        lang_name = self.translate_language(macr['LCODE'])
        if country in self._country:
            country_name = gettext.dgettext('iso_3166', self._country[country])
            return country_name
        else:
            return country
#        if current_language:
#            os.environ["LANGUAGE"] = current_language
#        else:
#            del os.environ["LANGUAGE"]
#        return (lang_name, country_name)

    def translate(self, locale, native=False, allCountries=False):
        """ get a locale code and output a human readable name """
        returnVal = ''
        macr = macros.LangpackMacros(self._datadir, locale)
        if native == True:
            current_language = None
            if "LANGUAGE" in os.environ:
                current_language = os.environ["LANGUAGE"]
            os.environ["LANGUAGE"] = macr["LOCALE"]

        lang_name = self.translate_language(macr["LCODE"])
        returnVal = lang_name
        if len(macr["CCODE"]) > 0:
            country_name = self.translate_country(macr["CCODE"])
            # get all locales for this language
            l = [k for k in self.generated_locales() if k.startswith(macr['LCODE'])]
            # only show region/country if we have more than one 
            if (allCountries == False and len(l) > 1) or allCountries == True:
                mycountry = self.country(macr['CCODE'])
                if mycountry:
                    returnVal = "%s (%s)" % (lang_name, country_name)
        if len(macr["VARIANT"]) > 0:
            returnVal = "%s - %s" % (returnVal, macr["VARIANT"])
        
        if native == True:
            if current_language:
                os.environ["LANGUAGE"] = current_language
            else:
                del os.environ["LANGUAGE"]
        return returnVal
         
#        if "_" in locale:
#            #(lang, country) = locale.split("_")
#            (lang_name, country_name) = self.translate_locale(locale)
#            # get all locales for this language
#            l = [k for k in self.generated_locales() if k.startswith(macr['LCODE'])]
#            # only show region/country if we have more than one 
#            if len(l) > 1:
#                mycountry = self.country(macr['CCODE'])
#                if mycountry:
#                    return "%s (%s)" % (lang_name, country_name)
#                else:
#                    return lang_name
#            else:
#                return lang_name
#        return self.translate_language(locale)

    def makeEnvString(self, code):
        """ input is a language code, output a string that can be put in
            the LANGUAGE enviroment variable.
            E.g: en_DK -> en_DK:en
        """
        if not code:
            return ''
        macr = macros.LangpackMacros(self._datadir, code)
        langcode = macr['LCODE']
        locale = macr['LOCALE']
        # first check if we got somethign from languagelist
        if locale in self._languagelist:
            langlist = self._languagelist[locale]
        # if not, fall back to "dumb" behaviour
        elif locale == langcode:
            langlist = locale
        else:
            langlist = "%s:%s" % (locale, langcode)
        if not (langlist.endswith(':en') or langlist == 'en'):
            langlist = "%s:en" % langlist
        return langlist

    def getUserDefaultLanguage(self):
        formats = ''
        language = ''
        result = []
        fname = os.path.expanduser("~/.pam_environment")
        if os.path.exists(fname) and \
           os.access(fname, os.R_OK):
            with open(fname) as f:
                for line in f:
                    match_language = re.match(r'LANGUAGE=(.*)$',line)
                    if match_language:
                        language = match_language.group(1)
        user_name = pwd.getpwuid(os.geteuid()).pw_name
        try:
            bus = dbus.SystemBus()
            obj = bus.get_object('org.freedesktop.Accounts', '/org/freedesktop/Accounts')
            iface = dbus.Interface(obj, dbus_interface='org.freedesktop.Accounts')
            user_path = iface.FindUserByName(user_name)

            obj = bus.get_object('org.freedesktop.Accounts', user_path)
            iface = dbus.Interface(obj, dbus_interface='org.freedesktop.DBus.Properties')
            formats = iface.Get('org.freedesktop.Accounts.User', 'FormatsLocale')
            if len(language) == 0:
                firstLanguage = iface.Get('org.freedesktop.Accounts.User', 'Language')
                language = self.makeEnvString(firstLanguage)
        except Exception as msg:
            # a failure here shouldn't trigger a fatal error
            warnings.warn(msg.args[0])
            pass
        if len(language) == 0 and "LANGUAGE" in os.environ:
            language = os.environ["LANGUAGE"]
        if len(formats) == 0 and "LC_NAME" in os.environ:
            formats = os.environ["LC_NAME"]
        if len(formats) == 0 and "LANG" in os.environ:
            formats = os.environ["LANG"]
        if len(formats) > 0 and len(language) == 0:
            language = self.makeEnvString(formats)
        result.append(formats)
        result.append(language)
        return result

    def getSystemDefaultLanguage(self):
        lang = ''
        formats = ''
        language = ''
        result = []
        for fname in self.environments:
            if os.path.exists(fname) and \
               os.access(fname, os.R_OK):
                with open(fname) as f:
                    for line in f:
                        # support both LANG="foo" and LANG=foo
                        if line.startswith("LANG"):
                            line = line.replace('"','')
                        match_lang = re.match(r'LANG=(.*)$',line)
                        if match_lang:
                            lang = match_lang.group(1)
                        if line.startswith("LC_TIME"):
                            line = line.replace('"','')
                        match_formats = re.match(r'LC_TIME=(.*)$',line)
                        if match_formats:
                            formats = match_formats.group(1)
                        if line.startswith("LANGUAGE"):
                            line = line.replace('"','')
                        match_language = re.match(r'LANGUAGE=(.*)$',line)
                        if match_language:
                            language = match_language.group(1)
        if len(lang) == 0:
            # fall back is 'en_US'
            lang = 'en_US.UTF-8'
        if len(language) == 0:
            # LANGUAGE has not been defined, generate a string from the provided LANG value
            language = self.makeEnvString(lang)
        if len(formats) == 0:
            formats = lang
        result.append(formats)
        result.append(language)
        return result

    def isSetSystemFormats(self):
        if not os.access(self.environments[0], os.R_OK):
            return False
        with open(self.environments[0]) as f:
            for line in f:
                if line.startswith("LC_TIME="):
                    return True
        return False


if __name__ == "__main__":
    datadir = "/usr/share/language-selector/"
    li = LocaleInfo("languagelist", datadir)

    print("default system locale and languages: '%s'" % li.getSystemDefaultLanguage())
    print("default user locale and languages: '%s'" % li.getUserDefaultLanguage())

    print(li._lang)
    print(li._country)
    print(li._languagelist)
    print(li.generated_locales())



#!/usr/bin/python3

import unittest
import sys
import apt
import apt_pkg
sys.path.insert(0, "../")

from LanguageSelector.LanguageSelector import LanguageSelectorBase

class TestLanguageSelector(unittest.TestCase):

    def test_package_lists_good(self):
        " test for non networked sources "
        apt_pkg.config.set("Dir::State::lists","./test-data/var/lib/apt/lists")
        apt_pkg.config.set("Dir::State::status","./test-data/empty")
        apt_pkg.config.set("Dir::Etc::SourceList","./test-data/etc/apt/sources.list.good")
        apt_pkg.config.set("Dir::Etc::SourceParts","./xxx")
        ls = LanguageSelectorBase(datadir="../")
        ls.openCache(apt.progress.base.OpProgress())
        self.assertTrue(ls._cache.havePackageLists == True,
                      "verifyPackageLists returned False for a good list")

    def test_package_lists_fail(self):
        " test for non networked sources "
        apt_pkg.config.set("Dir::State::lists","./test-data/var/lib/apt/lists")
        apt_pkg.config.set("Dir::State::status","./test-data/empty")
        apt_pkg.config.set("Dir::Etc::SourceList","./test-data/etc/apt/sources.list.fail")
        ls = LanguageSelectorBase(datadir="../")
        ls.openCache(apt.progress.base.OpProgress())
        self.assertTrue(ls._cache.havePackageLists == False,
                      "verifyPackageLists returned True for a list with missing indexfiles")
        


if __name__ == "__main__":
    apt_pkg.config.set("Apt::Architecture","i386")
    unittest.main()



#!/usr/bin/python3

import apt
import subprocess

DEFAULT_DEPENDS_FILE='/usr/share/language-selector/data/pkg_depends'

class LanguageSupport:
    lang_country_map = None

    def __init__(self, apt_cache=None, depends_file=None):
        if apt_cache is None:
            self.apt_cache = apt.Cache()
        else:
            self.apt_cache = apt_cache

        self.pkg_depends = self._parse_pkg_depends(depends_file or
                DEFAULT_DEPENDS_FILE)

    def by_package_and_locale(self, package, locale, installed=False):
        '''Get language support packages for a package and locale.

        Note that this does not include support packages which are not specific
        to a particular trigger package, e. g. general language packs. To get
        those, call this with package==''.

        By default, only return packages which are not installed. If installed
        is True, return all packages instead.
        '''
        packages = []
        depmap = self.pkg_depends.get(package, {})

        # check explicit entries for that locale
        for pkglist in depmap.get(self._langcode_from_locale(locale), {}).values():
            for p in pkglist:
                if p in self.apt_cache:
                    packages.append(p)

        # check patterns for empty locale string (i. e. applies to any locale)
        for pattern_list in depmap.get('', {}).values():
            for pattern in pattern_list:
                for pkg_candidate in self._expand_pkg_pattern(pattern, locale):
                    if pkg_candidate in self.apt_cache:
                        packages.append(pkg_candidate)

        if not installed:
            # filter out installed packages
            packages = [p for p in packages if not self.apt_cache[p].installed]
        return packages

    def by_locale(self, locale, installed=False):
        '''Get language support packages for a locale.

        Return all packages which need to be installed in order to provide
        language support for the given locale for all already installed
        packages. This should be called after adding a new locale to the
        system.

        By default, only return packages which are not installed. If installed
        is True, return all packages instead.
        '''
        packages = []

        for trigger in self.pkg_depends:
            try:
                if trigger == '' or self.apt_cache[trigger].installed:
                    packages += self.by_package_and_locale(trigger, locale, installed)
            except KeyError:
                continue

        return packages

    def by_package(self, package, installed=False):
        '''Get language support packages for a package.

        This will install language support for that package for all available
        system languages. This is a wrapper around available_languages() and
        by_package_and_locale().

        Note that this does not include support packages which are not specific
        to a particular trigger package, e. g. general language packs. To get
        those, call this with package==''.

        By default, only return packages which are not installed. If installed
        is True, return all packages instead.
        '''
        packages = set()
        for lang in self.available_languages():
            packages.update(self.by_package_and_locale(package, lang, installed))
        return packages

    def missing(self, installed=False):
        '''Get language support packages for current system.

        Return all packages which need to be installed in order to provide
        language support all system locales for all already installed
        packages. This should be called after installing the system without
        language support packages (perhaps because there was no network
        available to download them).

        This is a wrapper around available_languages() and by_locale().

        By default, only return packages which are not installed. If installed
        is True, return all packages instead.
        '''
        packages = set()
        for lang in self.available_languages():
                packages.update(self.by_locale(lang, installed))
        return self._hunspell_frami_special(packages)

    def _hunspell_frami_special(self, packages):
        ''' Ignore missing hunspell-de-xx if hunspell-de-xx-frami is installed.

        https://launchpad.net/bugs/1103547
        '''
        framis = []
        for country in ['de', 'at', 'ch']:
            frami = 'hunspell-de-' + country + '-frami'
            try:
                if self.apt_cache[frami].installed:
                    framis.append(frami)
            except KeyError:
                pass
        if len(framis) == 0:
            return packages
        packages_new = set()
        for pack in packages:
            if pack + '-frami' not in framis:
                packages_new.add(pack)
        return packages_new

    def available_languages(self):
        '''List available languages in the system.

        The list items can be passed as the "locale" argument of by_locale(),
        by_package_and_locale(), etc.
        '''
        languages = set()

        lang_string = subprocess.check_output(
            ['/usr/share/language-tools/language-options'],
            universal_newlines=True)

        for lang in lang_string.split():
            languages.add(lang)
            if not lang.startswith('zh_'):
                languages.add(lang.split('_')[0])

        return languages

    def _parse_pkg_depends(self, filename):
        '''Parse pkg_depends file.

        Return trigger_package -> langcode -> category -> [dependency,...] map.
        '''
        map = {}
        with open(filename) as f:
            for line in f:
                line = line.strip()
                if not line or line.startswith('#'):
                    continue

                (cat, lc, trigger, dep) = line.split(':')
                map.setdefault(trigger, {}).setdefault(lc, {}).setdefault(cat,
                        []).append(dep)

        return map

    @classmethod
    def _langcode_from_locale(klass, locale):
        '''Turn a locale name into a language code as in pkg_depends.'''

        # special-case Chinese locales, as they are split between -hans and
        # -hant
        if locale.startswith('zh_CN') or locale.startswith('zh_SG'):
            return 'zh-hans'
        # Hong Kong and Taiwan use traditional
        if locale.startswith('zh_'):
            return 'zh-hant'

        return locale.split('_', 1)[0]

    @classmethod
    def _expand_pkg_pattern(klass, pattern, locale):
        '''Return all possible suffixes for given pattern and locale'''

        # people might call this with the pseudo-locales "zh-han[st]", support
        # these as well; we can only guess the country here.
        if locale == 'zh-hans':
            locale = 'zh_CN'
        elif locale == 'zh-hant':
            locale = 'zh_TW'

        locale = locale.split('.', 1)[0].lower()
        variant = None
        country = None
        try:
            (lang, country) = locale.split('_', 1)
            if '@' in country:
                (country, variant) = country.split('@', 1)
        except ValueError:
            lang = locale

        pkgs = [pattern,
                '%s%s' % (pattern, lang)]

        if country:
            pkgs.append('%s%s%s' % (pattern, lang, country))
            pkgs.append('%s%s-%s' % (pattern, lang, country))
        else:
            for country in klass._countries_for_lang(lang):
                pkgs.append('%s%s%s' % (pattern, lang, country))
                pkgs.append('%s%s-%s' % (pattern, lang, country))

        if variant:
            pkgs.append('%s%s-%s' % (pattern, lang, variant))

        if country and variant:
            pkgs.append('%s%s-%s-%s' % (pattern, lang, country, variant))

        # special-case Chinese
        if lang == 'zh':
            if country in ['cn', 'sg']:
                pkgs.append(pattern + 'zh-hans')
            else:
                pkgs.append(pattern + 'zh-hant')

        return pkgs

    @classmethod
    def _countries_for_lang(klass, lang):
        '''Return a list of countries for given language'''

        if klass.lang_country_map is None:
            klass.lang_country_map = {}
            # fill cache
            with open('/usr/share/i18n/SUPPORTED') as f:
                for line in f:
                    line = line.split('#', 1)[0].split(' ')[0]
                    if not line:
                        continue
                    line = line.split('.', 1)[0].split('@')[0]
                    try:
                        (l, c) = line.split('_')
                    except ValueError:
                        continue
                    c = c.lower()
                    klass.lang_country_map.setdefault(l, set()).add(c)

        return klass.lang_country_map.get(lang, [])

def apt_cache_add_language_packs(resolver, cache, depends_file=None):
    '''Add language support for packages marked for installation.
    
    For all packages which are marked for installation in the given apt.Cache()
    object, mark the corresponding language packs and support packages for
    installation as well.

    This function can be used as an aptdaemon modify_cache_after plugin.
    '''
    ls = LanguageSupport(cache, depends_file)
    support_pkgs = set()
    for pkg in cache.get_changes():
        if pkg.marked_install:
            support_pkgs.update(ls.by_package(pkg.name))

    for pkg in support_pkgs:
        cache[pkg].mark_install(from_user=False)

def packagekit_what_provides_locale(cache, type, search, depends_file=None):
    '''PackageKit WhatProvides plugin for locale().'''

    if not search.startswith('locale('):
        raise NotImplementedError('cannot handle query type ' + search)

    locale = search.split('(', 1)[1][:-1]
    ls = LanguageSupport(cache, depends_file)
    pkgs = ls.by_locale(locale, installed=True)
    return [cache[p] for p in pkgs]




# (c) 2006 Canonical
# Author: Michael Vogt <michael.vogt@ubuntu.com>
#
# Released under the GPL
#

import os
import subprocess
import tempfile

def find_string_and_replace(findString, setString, file_list, 
                            startswith=True, append=True):
    """ find all strings that startswith findString and replace them with
        setString
    """
    for fname in file_list:
        out = tempfile.NamedTemporaryFile(delete=False,
                                          dir=os.path.dirname(fname))
        foundString = False
        if (os.path.exists(fname) and
            os.access(fname, os.R_OK)):
            # look for the line
            with open(fname) as f:
                for line in f:
                    tmp = line.strip()
                    if startswith and tmp.startswith(findString):
                        foundString = True
                        line = setString
                    if not startswith and tmp == findString:
                        foundString = True
                        line = setString
                    out.write(line.encode('UTF-8'))
        # if we have not found them append them
        if not foundString and append:
            out.write(setString.encode('UTF-8'))
        out.flush()
        # rename is atomic
        os.rename(out.name, fname)
        os.chmod(fname, 0o644)

def language2locale(language):
    """ generate locale name for LC_* environment variables
    """
    first_elem = language.split(':')[0]
    locale = subprocess.check_output(
        ['/usr/share/language-tools/language2locale', first_elem],
        universal_newlines=True)
    return locale.rstrip()




#!/usr/bin/python3

import unittest
import sys
sys.path.insert(0, "../")

from LanguageSelector.utils import *

class TestFindReplace(unittest.TestCase):

    def setUp(self):
        self.test_string="""xxx
line
search_text and more
line 17
"""

    def test_find_replace(self):
        with open("foo.txt","w") as f:
            f.write(self.test_string)
        find_string_and_replace("search_text", "lala", ["foo.txt"])
        with open("foo.txt") as f:
            new_content = f.read()
        self.assertFalse("search_text" in new_content)
        self.assertTrue("lala" in new_content)
        os.unlink("foo.txt")

    def test_find_replace_not_in_file(self):
        with open("foo.txt","w") as f:
            f.write(self.test_string)
        find_string_and_replace("string_not_in_file", "lala", ["foo.txt"])
        with open("foo.txt") as f:
            new_content = f.read()
        self.assertTrue("lala" in new_content)
        os.unlink("foo.txt")

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python3

from setuptools import setup
from DistUtilsExtra.command import (build_extra, build_i18n, build_help,
                                    build_icons)
import os
import sys

setup(name='language-selector',
      version='0.1',
      py_modules = ['language_support_pkgs'],
      packages=['LanguageSelector',
                'LanguageSelector.gtk'],
      scripts=['gnome-language-selector',
               'check-language-support'],
      data_files=[('share/language-selector/data',
                   ["data/language-selector.png",
                    "data/languagelist",
                    "data/langcode2locale",
                    "data/locale2langpack",
                    "data/pkg_depends",
                    "data/variants",
                    "data/LanguageSelector.ui"]),
                  # dbus stuff
                  ('share/dbus-1/system-services',
                   ['dbus_backend/com.ubuntu.LanguageSelector.service']),
                  ('../etc/dbus-1/system.d/',
                   ["dbus_backend/com.ubuntu.LanguageSelector.conf"]),
                  ('lib/language-selector/',
                   ["dbus_backend/ls-dbus-backend"]),
                  # pretty pictures
                  ('share/pixmaps',
                   ["data/language-selector.png"]),
                  ],
      entry_points='''[aptdaemon.plugins]
modify_cache_after=language_support_pkgs:apt_cache_add_language_packs
[packagekit.apt.plugins]
what_provides=language_support_pkgs:packagekit_what_provides_locale
''',
      cmdclass={"build": build_extra.build_extra,
                "build_i18n": build_i18n.build_i18n,
                "build_help": build_help.build_help,
                "build_icons": build_icons.build_icons,
                },

      )




from __future__ import print_function

import warnings
warnings.filterwarnings("ignore", "apt API not stable yet", FutureWarning)
import apt

import language_support_pkgs

class LanguagePackageStatus(object):
    def __init__(self, languageCode, pkg_template):
        self.languageCode = languageCode
        self.pkgname_template = pkg_template
        self.available = False
        self.installed = False
        self.doChange = False

    def __str__(self):
        return 'LanguagePackageStatus(langcode: %s, pkgname %s, available: %s, installed: %s, doChange: %s' % (
                self.languageCode, self.pkgname_template, str(self.available),
                str(self.installed), str(self.doChange))

# the language-support information
class LanguageInformation(object):
    def __init__(self, cache, languageCode=None, language=None):
        #FIXME:
        #needs a new structure:
        #languagePkgList[LANGCODE][tr|fn|in|wa]=[packages available for that language in that category]
        #@property for each category
        #@property for each LANGCODE
        self.languageCode = languageCode
        self.language = language
        # langPack/support status 
        self.languagePkgList = {}
        self.languagePkgList["languagePack"] = LanguagePackageStatus(languageCode, "language-pack-%s")
        for langpkg_status in self.languagePkgList.values():
            pkgname = langpkg_status.pkgname_template % languageCode
            langpkg_status.available = pkgname in cache
            if langpkg_status.available:
                langpkg_status.installed = cache[pkgname].is_installed
        
    @property
    def inconsistent(self):
        " returns True if only parts of the language support packages are installed "
        if (not self.notInstalled and not self.fullInstalled) : return True
        return False
    @property
    def fullInstalled(self):
        " return True if all of the available language support packages are installed "
        for pkg in self.languagePkgList.values() :
            if not pkg.available : continue
            if not ((pkg.installed and not pkg.doChange) or (not pkg.installed and pkg.doChange)) : return False
        return True
    @property
    def notInstalled(self):
        " return True if none of the available language support packages are installed "
        for pkg in self.languagePkgList.values() :
            if not pkg.available : continue
            if not ((not pkg.installed and not pkg.doChange) or (pkg.installed and pkg.doChange)) : return False
        return True
    @property
    def changes(self):
        " returns true if anything in the state of the language packs/support changes "
        for pkg in self.languagePkgList.values() :
            if (pkg.doChange) : return True
        return False
    def __str__(self):
        return "%s (%s)" % (self.language, self.languageCode)

# the pkgcache stuff
class ExceptionPkgCacheBroken(Exception):
    pass

class LanguageSelectorPkgCache(apt.Cache):

    def __init__(self, localeinfo, progress):
        apt.Cache.__init__(self, progress)
        if self._depcache.broken_count > 0:
            raise ExceptionPkgCacheBroken()
        self._localeinfo = localeinfo
        self.lang_support = language_support_pkgs.LanguageSupport(self)

    @property
    def havePackageLists(self):
        " verify that a network package lists exists "
        for metaindex in self._list.list:
            for indexfile in metaindex.index_files:
                if indexfile.archive_uri("").startswith("cdrom:"):
                    continue
                if indexfile.archive_uri("").startswith("http://security.ubuntu.com"):
                    continue
                if indexfile.label != "Debian Package Index":
                    continue
                if indexfile.exists and indexfile.has_packages:
                    return True
        return False

    def clear(self):
        """ clear the selections """
        self._depcache.init()

    def getChangesList(self):
        to_inst = []
        to_rm = []
        for pkg in self.get_changes():
            if pkg.marked_install or pkg.marked_upgrade:
                to_inst.append(pkg.name)
            if pkg.marked_delete:
                to_rm.append(pkg.name)
        return (to_inst,to_rm)

    def tryChangeDetails(self, li):
        " commit changed status of list items"""
        # we iterate over items of type LanguagePackageStatus
        for (key, item) in li.languagePkgList.items():
            if item.doChange:
                pkgs = self.lang_support.by_locale(li.languageCode, installed=item.installed)
                #print("XXX pkg list for lang %s, installed: %s" % (item.languageCode, str(item.installed)))
                try:
                    if item.installed:
                        # We are selective when deleting language support packages to
                        # prevent removal of packages that are not language specific.
                        for pkgname in pkgs:
                            if pkgname.startswith('language-pack-') or \
                               pkgname.endswith('-' + li.languageCode):
                                self[pkgname].mark_delete()
                    else:
                        for pkgname in pkgs:
                            self[pkgname].mark_install()
                except SystemError:
                    raise ExceptionPkgCacheBroken()

    def getLanguageInformation(self):
        """ returns a list with language packs/support packages """
        res = []
        for (code, lang) in self._localeinfo._lang.items():
            if code == 'zh':
                continue
            li = LanguageInformation(self, code, lang)
            if [s for s in li.languagePkgList.values() if s.available]:
                res.append(li)

        return res


if __name__ == "__main__":

    from LocaleInfo import LocaleInfo
    datadir = "/usr/share/language-selector"
    li = LocaleInfo("languagelist", datadir)

    lc = LanguageSelectorPkgCache(li,apt.progress.OpProgress())
    print("available language information")
    print(", ".join(["%s" %x for x in lc.getLanguageInformation()]))



from __future__ import print_function

import libxml2

class Variant:
        def __init__(self, name, desc, raw_desc):
                self.name = name
                self.desc = desc
                self.raw_desc = raw_desc

        def __str__(self):
                return "%s: %s, %s" % (self.name, self.desc, self.raw_desc)

class Layout:
        def __init__(self, name, desc, raw_desc, short_desc, raw_short_desc, variants):
                self.name = name
                self.desc = desc
                self.raw_desc = raw_desc
                self.short_desc = short_desc
                self.raw_short_desc = raw_short_desc
                self.variants = variants
                
        def __str__(self):
                return "%s: %s, %s; %s, %s;; %s" % (self.name,self.desc,self.raw_desc,self.short_desc,self.raw_short_desc,["%s" % x for x in self.variants])

def get_all_layout_possibilities():
        possibility_list = list()
        
        #FIXME: don't call parseFile() twice
        doc = libxml2.parseFile("/etc/X11/xkb/rules/xorg.xml")
        ctxt = doc.xpathNewContext()
                        
        for i in ctxt.xpathEval("/xkbConfigRegistry/layoutList/layout/configItem/name/text()"):
                possibility_list.append(i.content)
        
        return possibility_list

def get_variants(layout_node, lang):
        variant_list = list()

        variant_nodes = layout_node.xpathEval("../../../variantList/variant/configItem/name/text()")
        for i in variant_nodes:
                if len(i.xpathEval("../description[@xml:lang='%s']" % lang)) > 0:
                        trans = i.xpathEval("../description[@xml:lang='%s']" % lang)[0]
                else:
                        trans = ""

                v = Variant(i.content, trans, i.xpathEval("../../description[position()=1]")[0].content)
                variant_list.append(v)
        
        return variant_list


def get_layouts(lang):
        layout_list = list()
        doc = libxml2.parseFile("/etc/X11/xkb/rules/xorg.xml")
        ctxt = doc.xpathNewContext()
        layout_nodes = ctxt.xpathEval("/xkbConfigRegistry/layoutList/layout/configItem/name/text()")

        for i in layout_nodes:
                if i.content == lang:
                        if (len(i.xpathEval("../description[@xml:lang='%s']" % lang)) > 0):
                                translation = i.xpathEval("../description[@xml:lang='%s']" % lang)[0]
                        else:
                                translation = ""
                        if (len(i.xpathEval("../description[@xml:lang='%s']" % lang)) > 0):
                                short_trans = i.xpathEval("../shortDescription[@xml:lang='%s']" % lang)[0]
                        else:
                                short_trans = ""

                        layout_list.append(Layout(i.content, 
                                translation, i.xpathEval("../../description[position()=1]")[0].content, 
                                short_trans, i.xpathEval("../../shortDescription[position()=1]")[0].content,
                                get_variants(i, lang)))

        return layout_list


if __name__ == "__main__":
        for i in get_layouts("fr"): 
                print(i)

        for i in get_all_layout_possibilities():
                print(i)
